/****************************************************************************/
/* SUMMARY:                                                                 */
/* The script loads Oracle optimizer statistic collection stored procedure  */
/* into the current schema.                                                 */
/* Stored procedure can be used to collect statistics on BLOM database or   */
/* emit a list of PL/SQL commands that can be used to do the same thing.    */
/*                                                                          */
/* Current logic is as follows:                                             */
/* - if based on current statistics a table has more than p_threshold_rows  */
/*   records, statistics will be estimated using 25% row sampling of the    */
/*   total number of rows;                                                  */
/* - if based on current statistics a table has p_threshold_rows records or */
/*   less, statistics will be computed (estimated using 100% row sampling); */
/* - if table has no statistics collected, decision to use 25% estimate or  */
/*   compute will be made on total size of the table segments (index        */
/*   segments are not taken into account):                                  */
/*   -- if size of the table segments in bytes is more                      */
/*      than p_threshold_bytes, statistics will be estimated using 25% row  */
/*      sampling;                                                           */
/*   -- otherwise statistics will be computed.                              */
/*                                                                          */
/* STORED PROCEDURE INFO                                                    */
/*                                                                          */
/* INPUT PARAMERES:                                                         */
/* p_threshold_rows  - threshold based on the number of rows, the default   */
/*                     is one million rows                                  */
/* p_threshold_bytes - threshold based on the table size, the default is    */
/*                     one hundred mega bytes                               */
/* p_no_invalidate   - indicates whether to invalidate the child cursors    */
/*                     currently cached in the SGA                          */
/* p_print_only      - do not gather statistics, but emit PL/SQL commands   */
/*                     for each tables instead (requires serveroutput to be */
/*                     turned on on the client and set to unlimited size)   */
/*                                                                          */
/* OUTPUT PARAMERES:                                                        */
/* None                                                                     */
/****************************************************************************/
CREATE OR REPLACE PROCEDURE
BL_GATHER_SCHEMA_STATS
(
 p_threshold_rows  IN INTEGER  := 1000000,
 p_threshold_bytes IN INTEGER  := 104857600,
 p_no_invalidate   IN VARCHAR2 := 'Y',
 p_print_only      IN VARCHAR2 := 'N'
)
AS
  TYPE stmt_list_t IS TABLE OF VARCHAR2(4000) INDEX BY PLS_INTEGER;
  lsc_pct_below VARCHAR2(10) := '100';
  lsc_pct_above VARCHAR2(10) := '25';
  lsc_num_fmt   VARCHAR2(30) := 'TM9';
  lsc_dte_fmt   VARCHAR2(30) := 'YYYY-MM-DD HH24:MI:SS';
  lsc_rp_tbl    VARCHAR2(30) := ':table_name';
  lsc_rp_pct    VARCHAR2(30) := ':estimate_percent';
  lsc_rp_ninv   VARCHAR2(30) := ':no_invalidate';
  lsc_rp_owner  VARCHAR2(30) := ':schema_owner';
  lsc_col_pfx   VARCHAR2(30) := 'begin ';
  lsc_col_sfx   VARCHAR2(30) := '; end;';
  lsc_prt_pfx   VARCHAR2(30) := 'exec ';
  lsc_prt_sfx   VARCHAR2(30) := ';';
  lsc_stmt_tpl  VARCHAR2(4000) := 
    'dbms_stats.gather_table_stats(' ||
    'estimate_percent=>' || lsc_rp_pct ||
    ',tabname=>''' || lsc_rp_tbl || '''' ||
    ',ownname=>''' || lsc_rp_owner || '''' ||
    ',cascade=>true' ||
    ',degree=>DBMS_STATS.AUTO_DEGREE' ||
    ',no_invalidate=>' || lsc_rp_ninv ||
    ',force=>true' ||
    ',method_opt=>''for all columns size 254''' ||
    ')';
  --
  lv_exec_stmts  stmt_list_t;
  lv_thresh_rws  INTEGER := NVL(p_threshold_rows, 1000000);
  lv_thresh_bts  INTEGER := NVL(p_threshold_bytes, 104857600);
  lv_curr_schema VARCHAR2(30) := SYS_CONTEXT('USERENV', 'CURRENT_SCHEMA');
  lv_print_only  BOOLEAN := FALSE;
  lv_stmt_tpl    VARCHAR2(4000);
  lv_below_tpl   VARCHAR2(4000);
  lv_above_tpl   VARCHAR2(4000);
  lv_prefix      VARCHAR2(30) := '';
  lv_suffix      VARCHAR2(30) := '';
  lv_cnt         PLS_INTEGER;
  lv_gen_date    DATE := SYSDATE;
  --
  PROCEDURE do_preprocessing
  AS
  BEGIN
    lv_print_only := NVL(p_print_only, 'N') = 'Y';
    lv_stmt_tpl := REPLACE(lsc_stmt_tpl, lsc_rp_ninv, 
                     CASE WHEN NVL(p_no_invalidate, 'Y') = 'N' THEN
                            'false'
                          ELSE
                            'true'
                     END);
    lv_stmt_tpl := REPLACE(lv_stmt_tpl, lsc_rp_owner, lv_curr_schema);
    --
    IF lv_print_only THEN
      lv_prefix := lsc_prt_pfx;
      lv_suffix := lsc_prt_sfx;
    ELSE
      lv_prefix := lsc_col_pfx;
      lv_suffix := lsc_col_sfx;
    END IF;
    lv_below_tpl  := REPLACE(lv_stmt_tpl, lsc_rp_pct, lsc_pct_below);
    lv_above_tpl  := REPLACE(lv_stmt_tpl, lsc_rp_pct, lsc_pct_above);
  END do_preprocessing;
  --
  PROCEDURE collect_statements
  AS
  BEGIN
    -- process tables with the number of records above threshold
    SELECT
      lv_prefix ||
      REPLACE
      (
       CASE WHEN t.num_rows is NULL THEN
              CASE WHEN (
                         SELECT
                                SUM(s.BYTES)
                         FROM
                                USER_SEGMENTS s
                         WHERE
                                s.SEGMENT_NAME = t.TABLE_NAME
                                and
                                s.SEGMENT_TYPE = 'TABLE'
                        ) > lv_thresh_bts THEN
                     lv_above_tpl
                   ELSE
                     lv_below_tpl
              END
            WHEN t.num_rows > lv_thresh_rws THEN
              lv_above_tpl
            ELSE
              lv_below_tpl
       END,
       lsc_rp_tbl,
       t.table_name
      ) ||
      lv_suffix stmt
      BULK COLLECT INTO lv_exec_stmts
    FROM
      user_tables t
    WHERE
      temporary = 'N'
    ORDER BY
      TRUNC(t.num_rows/lv_thresh_rws) NULLS LAST,
      t.num_rows,
      t.table_name
    ;
    --
    lv_cnt := lv_exec_stmts.COUNT();
    --
  END collect_statements;
  --
  PROCEDURE do_print_only
  AS
  BEGIN
    DBMS_OUTPUT.PUT_LINE('-- Gather stats script for Oracle database. ');
    DBMS_OUTPUT.PUT_LINE('-- Current schema is [' || USER || '].');
    DBMS_OUTPUT.PUT_LINE('-- Generated at [' || 
    to_char(lv_gen_date, lsc_dte_fmt) || '].');
    --
    IF lv_cnt > 0 THEN
      DBMS_OUTPUT.PUT_LINE('-- Table row counts are read from USER_TABLES view.');
      DBMS_OUTPUT.PUT_LINE('-- For tables with row count above ' || 
      to_char(lv_thresh_rws, lsc_num_fmt) || ', estimate_percent is ' ||
      lsc_pct_above || '.');
      DBMS_OUTPUT.PUT_LINE('-- For tables with row count ' || 
      to_char(lv_thresh_rws, lsc_num_fmt) || ' or less, estimate_percent is ' || 
      lsc_pct_below || '.');
      DBMS_OUTPUT.PUT_LINE('-- For tables that weren''t analyzed yet ' || 
      'estimate_percent is determined based on the total size of the table segments in bytes.');
      DBMS_OUTPUT.PUT_LINE('--   if total size in bytes is above ' || 
      to_char(lv_thresh_bts, lsc_num_fmt) || 'estimate_percent is ' || 
      lsc_pct_above || '.');
      DBMS_OUTPUT.PUT_LINE('--   if total size in bytes is ' || 
      to_char(lv_thresh_bts, lsc_num_fmt) || ' or less, estimate_percent is ' || 
      lsc_pct_below || '.');
      --
      DBMS_OUTPUT.PUT_LINE('exec dbms_output.put_line(''Started stats collection at '' ' ||
      '|| to_char(sysdate, ''yyyy-mm-dd hh24:mi:ss''));');
      FOR i IN 1..lv_cnt LOOP
        DBMS_OUTPUT.PUT_LINE(lv_exec_stmts(i));
      END LOOP;
      --
      DBMS_OUTPUT.PUT_LINE('exec dbms_output.put_line(''Finished stats collection at '' ' ||
      '|| to_char(sysdate, ''yyyy-mm-dd hh24:mi:ss''));');
      --
      DBMS_OUTPUT.PUT_LINE('-- Tables to gather stats on total: [' || 
      to_char(lv_cnt, lsc_num_fmt) || '].');
    ELSE
      DBMS_OUTPUT.PUT_LINE('-- No tables were found in the current schema.');
    END IF;
    --
  END do_print_only;
  --
  PROCEDURE do_gather_stats
  AS
  BEGIN
    IF lv_cnt > 0 THEN
      FOR i IN 1..lv_cnt LOOP
        -- DBMS_OUTPUT.PUT_LINE(lv_exec_stmts(i));
        EXECUTE IMMEDIATE lv_exec_stmts(i);
      END LOOP;
    END IF;
  END do_gather_stats;
  --
  PROCEDURE do_work
  AS
  BEGIN
    IF lv_print_only THEN
      do_print_only();
    ELSE
      do_gather_stats();
    END IF;
  END do_work;
BEGIN
  do_preprocessing();
  collect_statements();
  do_work();
END BL_GATHER_SCHEMA_STATS;
/
